# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys
import traceback
from pathlib import Path
from textwrap import dedent, indent
from typing import Dict, List, Optional, Sequence

import libcst as cst
from attr import dataclass, field

from .config import CAT_FIRST_PARTY, Config
from .util import stem_join, Timing, top_level_name

COMMENT_INDENT = "  "


def case_insensitive_ordering(text: Optional[str]) -> Optional[str]:
    if text is None:
        return None
    return str.casefold(text)


@dataclass
class Options:
    debug: bool


@dataclass
class SortWarning:
    line: int
    message: str


@dataclass
class Result:
    path: Path
    content: bytes = b""
    output: bytes = b""
    # encoding will be None on parse errors; we get this from LibCST on a successful
    # parse.
    encoding: Optional[str] = None
    error: Optional[Exception] = None
    trace: str = ""
    timings: Sequence[Timing] = ()
    warnings: Sequence[SortWarning] = ()

    def __attrs_post_init__(self) -> None:
        if self.error:
            exc_type, exc, tb = sys.exc_info()
            if exc_type is not None:
                self.trace = "".join(traceback.format_exception(exc_type, exc, tb))


@dataclass
class ImportItemComments:
    before: List[str] = field(factory=list)
    inline: List[str] = field(factory=list)
    following: List[str] = field(factory=list)

    def __add__(self, other: "ImportItemComments") -> "ImportItemComments":
        if not isinstance(other, ImportItemComments):
            return NotImplemented

        return ImportItemComments(
            before=[*self.before, *other.before],
            inline=[*self.inline, *other.inline],
            following=[*self.following, *other.following],
        )


@dataclass
class ImportComments:
    before: List[str] = field(factory=list)
    first_inline: List[str] = field(factory=list)
    initial: List[str] = field(factory=list)
    inline: List[str] = field(factory=list)  # Only when no trailing comma
    final: List[str] = field(factory=list)
    last_inline: List[str] = field(factory=list)

    def __add__(self, other: "ImportComments") -> "ImportComments":
        if not isinstance(other, ImportComments):
            return NotImplemented

        return ImportComments(
            before=[*self.before, *other.before],
            first_inline=[*self.first_inline, *other.first_inline],
            initial=[*self.initial, *other.initial],
            inline=[*self.inline, *other.inline],
            final=[*self.final, *other.final],
            last_inline=[*self.last_inline, *other.last_inline],
        )


@dataclass(order=True)
class SortKey:
    category_index: int
    is_from_import: bool
    ndots: int


@dataclass(order=True)
class SortableImportItem:
    name: str = field(order=str.casefold)
    asname: Optional[str] = field(eq=True, order=case_insensitive_ordering)
    comments: ImportItemComments = field(
        eq=False, order=False, factory=ImportItemComments
    )
    stem: Optional[str] = field(order=False, default=None, repr=False)
    comma: bool = field(eq=False, order=False, default=False, repr=False)

    @property
    def fullname(self) -> str:
        return stem_join(self.stem, self.name)

    def __add__(self, other: "SortableImportItem") -> "SortableImportItem":
        if not isinstance(other, SortableImportItem):
            return NotImplemented

        if self.name != other.name and self.asname != other.asname:
            raise ValueError("name and asname must match")

        return SortableImportItem(
            name=self.name,
            asname=self.asname,
            comments=self.comments + other.comments,
            stem=self.stem,
        )


@dataclass(order=True, repr=False)
class SortableImport:
    sort_key: SortKey = field(init=False)
    stem: Optional[str] = field(order=case_insensitive_ordering)  # "from" imports
    items: List[SortableImportItem] = field()
    comments: ImportComments = field(order=False, factory=ImportComments)
    indent: str = field(order=False, default="")
    config: Config = field(eq=False, order=False, factory=Config)

    # for cli/debugging only
    node: cst.CSTNode = field(eq=False, order=False, factory=cst.EmptyLine)

    def __repr__(self) -> str:
        items = indent(("\n".join(f"{item!r}," for item in self.items)), "        ")
        return (
            dedent(
                """
                    SortableImport(
                        # sort_key = {sort_key!r},
                        stem = {stem!r},
                        items = [
                    {items}
                        ],
                        comments = {comments!r},
                        indent = {indent!r},
                    )
                """
            )
            .strip()
            .format(
                sort_key=self.sort_key,
                stem=self.stem,
                items=items,
                comments=self.comments,
                indent=self.indent,
            )
        )

    def __add__(self, other: "SortableImport") -> "SortableImport":
        if not isinstance(other, SortableImport):
            return NotImplemented

        if self.sort_key != other.sort_key and self.stem != other.stem:
            raise ValueError("sort_key and stem must match")

        # Combine the items from the other import statement with items from this import.
        # If both the name and asname match, we can safely just add-op them to combine
        # comments, and assume the other import object will get discarded. If there is
        # no match, new items from other will be appended to the end of the list.
        #
        #   from foo import a, b, c
        #   from foo import b, c as C, d
        #
        # should get combined to
        #
        #   from foo import a, b, c, c as C, d
        #
        combined_items = list(self.items)
        for item in other.items:
            if item in combined_items:
                idx = combined_items.index(item)
                combined_items[idx] += item

            else:
                combined_items.append(item)

        return SortableImport(
            stem=self.stem,
            items=combined_items,
            comments=self.comments + other.comments,
            indent=self.indent,
            config=self.config,
            node=self.node,
        )

    @property
    def trailing_comma(self) -> bool:
        if self.items:
            return self.items[-1].comma
        return False

    @property
    def imported_names(self) -> Dict[str, str]:
        results: Dict[str, str] = {}

        for item in self.items:
            key = item.asname or item.name
            if self.stem is None and not item.asname:
                key = top_level_name(item.name)
                value = key
            else:
                value = item.fullname
            results[key] = value

        return results

    def calculate_sort_key(self) -> None:
        top: Optional[str] = None
        ndots = 0

        if self.stem is None:
            top = top_level_name(self.items[0].name)
        elif not self.stem.startswith("."):
            top = top_level_name(self.stem)
        else:
            # replicate ... sorting before .. before ., but after absolute
            ndots = 100 - (len(self.stem) - len(self.stem.lstrip(".")))

        category = self.config.category(top) if top else CAT_FIRST_PARTY

        self.sort_key = SortKey(
            # TODO this will raise on missing category
            category_index=self.config.categories.index(category),
            is_from_import=bool(self.stem),
            ndots=ndots,
        )

    def __attrs_post_init__(self) -> None:
        self.calculate_sort_key()


@dataclass(repr=False)
class SortableBlock:
    start_idx: int
    end_idx: int  # half-open interval

    imports: List[SortableImport] = field(factory=list)
    imported_names: Dict[str, str] = field(factory=dict)

    def __repr__(self) -> str:
        imports = indent("\n".join(f"{imp!r}," for imp in self.imports), "        ")
        return (
            dedent(
                """
                    SortableBlock(
                        start_idx = {start_idx!r},
                        end_idx = {end_idx!r},
                        imports = [
                    {imports}
                        ],
                        imported_names = {imported_names!r},
                    )
                """
            )
            .strip()
            .format(
                start_idx=self.start_idx,
                end_idx=self.end_idx,
                imports=imports,
                imported_names=self.imported_names,
            )
        )

    def add_import(self, imp: SortableImport, idx: int) -> None:
        self.end_idx = idx + 1
        self.imports.append(imp)
        for key, value in imp.imported_names.items():
            self.imported_names[key] = value



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
import unittest
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent
from typing import AnyStr, Generator
from unittest.mock import Mock, patch

import volatile
from click.testing import CliRunner

from ..cli import main


@contextmanager
def chdir(new_dir: str) -> Generator[None, None, None]:
    cur_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(cur_dir)


@contextmanager
def sample_contents(s: AnyStr) -> Generator[str, None, None]:
    with volatile.dir() as dtmp:
        ptmp = Path(dtmp)
        (ptmp / "pyproject.toml").write_text("")
        if isinstance(s, bytes):
            (ptmp / "sample.py").write_bytes(s)
        else:
            (ptmp / "sample.py").write_text(s)
        yield dtmp


class CliTest(unittest.TestCase):
    def test_benchmark(self) -> None:
        """Doesn't trigger multiprocessing (walk first in timings)"""
        with sample_contents("import sys\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["--benchmark", "check", "."])

        self.assertRegex(
            result.output,
            dedent(
                r"""
                walking \.:\s+\d+ µs
                parsing sample\.py:\s+\d+ µs
                sorting sample\.py:\s+\d+ µs
                total:\s+\d+ µs
                """
            ).strip(),
        )
        self.assertEqual(0, result.exit_code)

    def test_benchmark_multiple(self) -> None:
        "Does trigger multiprocessing (walk after parsing/sorting in timings)"
        with sample_contents("import sys\n") as dtmp:
            (Path(dtmp) / "foo.py").write_text("print('hello')\n")

            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["--benchmark", "check", "."])

        self.assertRegex(
            result.output,
            dedent(
                r"""
                parsing foo\.py:\s+\d+ µs
                sorting foo\.py:\s+\d+ µs
                """
            ).strip(),
        )

        self.assertRegex(
            result.output,
            dedent(
                r"""
                parsing sample\.py:\s+\d+ µs
                sorting sample\.py:\s+\d+ µs
                """
            ).strip(),
        )

        self.assertRegex(
            result.output,
            dedent(
                r"""
                walking \.:\s+\d+ µs
                total:\s+\d+ µs
                """
            ).strip(),
        )
        self.assertEqual(0, result.exit_code)

    @patch("usort.cli.sys")
    @patch("usort.cli.logging")
    def test_debug(self, mock_logging: Mock, mock_sys: Mock) -> None:
        mock_logging.DEBUG = logging.DEBUG
        mock_logging.WARNING = logging.WARNING
        mock_sys.stderr = object()

        with sample_contents("import foo\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                with self.subTest("no debug"):
                    result = runner.invoke(main, ["check", "."])
                    mock_logging.basicConfig.assert_called_with(
                        level=logging.WARNING, stream=mock_sys.stderr
                    )
                    self.assertEqual("", result.output)

                with self.subTest("with debug"):
                    result = runner.invoke(main, ["--debug", "check", "."])
                    mock_logging.basicConfig.assert_called_with(
                        level=logging.DEBUG, stream=mock_sys.stderr
                    )
                    self.assertEqual("", result.output)

    def test_check_no_change(self) -> None:
        with sample_contents("import sys\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["check", "."])

        self.assertEqual("", result.output)
        self.assertEqual(0, result.exit_code)

    def test_check_with_change(self) -> None:
        with sample_contents("import sys\nimport os\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["check", "."])

        self.assertEqual("Would sort sample.py\n", result.output)
        self.assertEqual(2, result.exit_code)

    def test_check_with_warnings(self) -> None:
        with sample_contents(b"import os\nimport sys\nimport re as os\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["check", "."])

        self.assertEqual(
            "Warning at sample.py:3 Name 'os' shadowed by 're'; implicit block split\n"
            "Would sort sample.py\n",
            result.output,
        )

    def test_diff_no_change(self) -> None:
        with sample_contents("import sys\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["diff", "."])

        self.assertEqual("", result.output)
        self.assertEqual(0, result.exit_code)

    def test_diff_with_change(self) -> None:
        with sample_contents(b"import sys\nimport os\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["diff", "."])

        self.assertEqual(
            """\
--- a/sample.py
+++ b/sample.py
@@ -1,2 +1,2 @@
+import os
 import sys
-import os
""".replace(
                "\r", ""
            ),
            result.output,
        )

        self.assertEqual(result.exit_code, 0)

    def test_list_imports(self) -> None:
        with sample_contents("import sys\nx = 5\nimport os") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                # TODO this takes filenames, not paths...
                result = runner.invoke(main, ["list-imports", "sample.py"])

        self.assertEqual(
            """\
sample.py 2 blocks:
  body[0:1]
Formatted:
[[[
import sys
]]]
  body[2:3]
Formatted:
[[[
import os
]]]
""",
            result.output.replace("\r\n", "\n"),
        )
        self.assertEqual(result.exit_code, 0)

    def test_format_no_change(self) -> None:
        with sample_contents("import sys\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["format", "."])

        self.assertEqual(result.output, "")
        self.assertEqual(result.exit_code, 0)

    def test_format_parse_error(self) -> None:
        """Code that has syntax that would never be valid in any version of python"""
        with sample_contents("import\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["format", "."])

        self.assertRegex(
            result.output,
            r"Error sorting sample\.py: Syntax Error @ 1:1\.\n"
            r"parser error: error at 2:0: expected NAME",
        )
        self.assertEqual(result.exit_code, 1)

    def test_format_parse_error_conflicting_syntax(self) -> None:
        """Code that contains syntax both <=2.7 and >=3.8 that could never coexist"""
        with sample_contents("while (i := foo()):\n    print 'i'\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["format", "."])

        self.assertRegex(
            result.output,
            r"Error sorting sample\.py: Syntax Error @ 2:1\.\n"
            r"parser error: error at 2:13: expected one of ",
        )
        self.assertEqual(result.exit_code, 1)

    def test_format_permission_error(self) -> None:
        """File does not have read permissions"""
        with sample_contents("print('hello world')\n") as dtmp:
            runner = CliRunner()
            # make the file unreadable
            (Path(dtmp) / "sample.py").chmod(0o000)
            with chdir(dtmp):
                result = runner.invoke(main, ["format", "."])
            # restore permissions so that cleanup can succeed on windows
            (Path(dtmp) / "sample.py").chmod(0o644)

        self.assertRegex(
            result.output,
            r"Error sorting sample\.py: .+ Permission denied:",
        )
        self.assertEqual(result.exit_code, 1)

    def test_format_with_change(self) -> None:
        with sample_contents("import sys\nimport os\n") as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["format", "."])

            self.assertEqual(result.output, "Sorted sample.py\n")
            self.assertEqual(result.exit_code, 0)

            self.assertEqual(
                """\
import os
import sys
""",
                (Path(dtmp) / "sample.py").read_text(),
            )

    def test_format_utf8(self) -> None:
        # the string is "µ" as in "µsort"
        with sample_contents(
            b"""\
import b
import a
s = "\xc2\xb5"
"""
        ) as dtmp:
            runner = CliRunner()
            with chdir(dtmp):
                result = runner.invoke(main, ["diff", "."])

            # Diff output is unicode
            self.assertEqual(
                result.output,
                """\
--- a/sample.py
+++ b/sample.py
@@ -1,3 +1,3 @@
+import a
 import b
-import a
 s = "\u00b5"
""",
                result.output,
            )

            with chdir(dtmp):
                result = runner.invoke(main, ["format", "."])

            self.assertEqual(
                b"""\
import a
import b
s = "\xc2\xb5"
""",
                (Path(dtmp) / "sample.py").read_bytes(),
            )

    def test_format_latin_1(self) -> None:
        # the string is "µ" as in "µsort"
        with sample_contents(
            b"""\
# -*- coding: latin-1 -*-
import b
import a
s = "\xb5"
""".replace(
                b"\r", b""
            )  # git on windows might make \r\n
        ) as dtmp:
            runner = CliRunner()

            # Diff output is unicode
            with chdir(dtmp):
                result = runner.invoke(main, ["diff", "."])

            self.assertEqual(
                result.output,
                """\
--- a/sample.py
+++ b/sample.py
@@ -1,4 +1,4 @@
 # -*- coding: latin-1 -*-
+import a
 import b
-import a
 s = "\u00b5"
""".replace(
                    "\r", ""
                ),  # git on windows again
                result.output,
            )

            # Format keeps current encoding
            with chdir(dtmp):
                result = runner.invoke(main, ["format", "."])

            self.assertEqual(
                b"""\
# -*- coding: latin-1 -*-
import a
import b
s = "\xb5"
""".replace(
                    b"\r", b""
                ),  # git on windows again
                (Path(dtmp) / "sample.py").read_bytes(),
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from pathlib import Path

import libcst as cst

from ..config import Config
from ..sorting import ImportSorter
from ..translate import import_from_node
from ..types import SortableImport, SortableImportItem
from ..util import parse_import


class SortableImportTest(unittest.TestCase):
    def test_import_from_node_Import(self) -> None:
        imp = import_from_node(parse_import("import a"), Config())
        self.assertIsNone(imp.stem)
        self.assertEqual("a", imp.items[0].name)
        self.assertEqual({"a": "a"}, imp.imported_names)

        imp = import_from_node(parse_import("import a, b"), Config())
        self.assertIsNone(imp.stem)
        self.assertEqual("a", imp.items[0].name)
        self.assertEqual("b", imp.items[1].name)
        self.assertEqual({"a": "a", "b": "b"}, imp.imported_names)

        imp = import_from_node(parse_import("import a as b"), Config())
        self.assertIsNone(imp.stem)
        self.assertEqual("a", imp.items[0].name)
        self.assertEqual("b", imp.items[0].asname)
        self.assertEqual({"b": "a"}, imp.imported_names)

        imp = import_from_node(parse_import("import os.path"), Config())
        self.assertIsNone(imp.stem)
        self.assertEqual("os.path", imp.items[0].name)
        self.assertEqual({"os": "os"}, imp.imported_names)

        imp = import_from_node(parse_import("import IPython.core"), Config())
        self.assertEqual("IPython.core", imp.items[0].name)
        self.assertEqual({"IPython": "IPython"}, imp.imported_names)

    def test_import_from_node_ImportFrom(self) -> None:
        imp = import_from_node(parse_import("from a import b"), Config())
        self.assertEqual("a", imp.stem)
        self.assertEqual("b", imp.items[0].name)
        self.assertEqual({"b": "a.b"}, imp.imported_names)

        imp = import_from_node(parse_import("from a import b as c"), Config())
        self.assertEqual("a", imp.stem)
        self.assertEqual("b", imp.items[0].name)
        self.assertEqual("c", imp.items[0].asname)
        self.assertEqual({"c": "a.b"}, imp.imported_names)

    def test_import_from_node_ImportFrom_relative(self) -> None:
        imp = import_from_node(parse_import("from .a import b"), Config())
        self.assertEqual(".a", imp.stem)
        self.assertEqual("b", imp.items[0].name)
        self.assertEqual({"b": ".a.b"}, imp.imported_names)

        imp = import_from_node(parse_import("from ...a import b"), Config())
        self.assertEqual("...a", imp.stem)
        self.assertEqual("b", imp.items[0].name)
        self.assertEqual({"b": "...a.b"}, imp.imported_names)

        imp = import_from_node(parse_import("from . import a"), Config())
        self.assertEqual(".", imp.stem)
        self.assertEqual("a", imp.items[0].name)
        self.assertEqual({"a": ".a"}, imp.imported_names)

        imp = import_from_node(parse_import("from .. import a"), Config())
        self.assertEqual("..", imp.stem)
        self.assertEqual("a", imp.items[0].name)
        self.assertEqual({"a": "..a"}, imp.imported_names)

        imp = import_from_node(parse_import("from . import a as b"), Config())
        self.assertEqual(".", imp.stem)
        self.assertEqual("a", imp.items[0].name)
        self.assertEqual("b", imp.items[0].asname)
        self.assertEqual({"b": ".a"}, imp.imported_names)

    def test_import_from_node_multiline(self) -> None:
        imp = import_from_node(parse_import("from a import (\n  b,\n  c\n)"), Config())
        expected = SortableImport(
            stem="a",
            items=[
                SortableImportItem(name="b", asname="", stem="a", comma=True),
                SortableImportItem(name="c", asname="", stem="a", comma=False),
            ],
        )
        self.assertListEqual(expected.items, imp.items)

        imp = import_from_node(parse_import("from a import (\n  b,\n  c,\n)"), Config())
        expected = SortableImport(
            stem="a",
            items=[
                SortableImportItem(name="b", asname="", stem="a", comma=True),
                SortableImportItem(name="c", asname="", stem="a", comma=True),
            ],
        )
        self.assertListEqual(expected.items, imp.items)


class IsSortableTest(unittest.TestCase):
    def test_is_sortable(self) -> None:
        sorter = ImportSorter(module=cst.Module([]), path=Path(), config=Config())
        self.assertTrue(sorter.is_sortable_import(parse_import("import a")))
        self.assertTrue(sorter.is_sortable_import(parse_import("from a import b")))
        self.assertFalse(
            sorter.is_sortable_import(parse_import("import a  # isort: skip"))
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import tempfile
import unittest
from pathlib import Path

from usort.config import CAT_FIRST_PARTY, CAT_FUTURE, CAT_THIRD_PARTY, Config
from .cli import chdir


class ConfigTest(unittest.TestCase):
    def test_third_party(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            # Defaults should not know about this module.
            # This is a silly value for default_category but this makes it more
            # obvious when it's in use.
            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
default_category = "future"
"""
            )
            conf = Config.find(Path(d))
            self.assertNotIn("psutil", conf.known)
            self.assertEqual(CAT_FUTURE, conf.category("psutil.cpu_times"))

            # "legacy" naming
            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
default_category = "future"
known_third_party = ["psutil", "cocoa"]
"""
            )

            conf = Config.find(Path(d))
            self.assertEqual(CAT_THIRD_PARTY, conf.known["cocoa"])
            self.assertEqual(CAT_THIRD_PARTY, conf.known["psutil"])
            self.assertEqual(CAT_THIRD_PARTY, conf.category("psutil.cpu_times"))

            # Works even on invalid children
            conf = Config.find(Path(d) / "foo")
            self.assertEqual(CAT_THIRD_PARTY, conf.known["cocoa"])
            self.assertEqual(CAT_THIRD_PARTY, conf.known["psutil"])
            self.assertEqual(CAT_THIRD_PARTY, conf.category("psutil.cpu_times"))

            # New naming
            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
default_category = "future"
[tool.usort.known]
third_party = ["psutil", "cocoa"]
"""
            )

            new_conf = Config.find(Path(d))
            self.assertEqual(conf, new_conf)

    def test_first_party_root_finding(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "a" / "b" / "c").mkdir(parents=True)
            (Path(d) / "a" / "b" / "__init__.py").write_text("")
            (Path(d) / "a" / "b" / "c" / "__init__.py").write_text("from b import zzz")

            f = Path(d) / "a" / "b" / "c" / "__init__.py"

            conf = Config.find(f)
            self.assertEqual(CAT_FIRST_PARTY, conf.known["b"])
            conf = Config.find(f.parent)  # c
            self.assertEqual(CAT_FIRST_PARTY, conf.known["b"])
            conf = Config.find(f.parent.parent)  # b
            self.assertEqual(CAT_FIRST_PARTY, conf.known["b"])
            conf = Config.find(Path("/"))
            self.assertNotIn("b", conf.known)

    def test_first_party_root_finding_disable(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "a" / "b" / "c").mkdir(parents=True)
            (Path(d) / "a" / "b" / "__init__.py").write_text("")
            (Path(d) / "a" / "b" / "c" / "__init__.py").write_text("from b import zzz")

            f = Path(d) / "a" / "b" / "c" / "__init__.py"

            conf = Config.find(f)
            self.assertEqual(CAT_FIRST_PARTY, conf.known["b"])

            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
first_party_detection = false
"""
            )

            conf = Config.find(f)
            self.assertNotIn("b", conf.known)

    def test_new_category_names(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort]
categories = ["future", "standard_library", "numpy", "third_party", "first_party"]
[tool.usort.known]
numpy = ["numpy", "pandas"]
"""
            )
            conf = Config.find(Path(d) / "sample.py")
            self.assertEqual("numpy", conf.known["pandas"])

    def test_new_category_names_invalid(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            (Path(d) / "pyproject.toml").write_text(
                """\
[tool.usort.known]
foo = ["numpy", "pandas"]
"""
            )
            with self.assertRaisesRegex(ValueError, "Known set for foo"):
                Config.find(Path(d) / "sample.py")

    def test_side_effect_init(self) -> None:
        config = Config()
        self.assertEqual([], config.side_effect_modules)
        self.assertEqual("", config.side_effect_re.pattern)
        self.assertRegex("", config.side_effect_re)

        config = Config(side_effect_modules=["fizzbuzz", "foo.bar.baz"])
        self.assertEqual(r"fizzbuzz\b|foo\.bar\.baz\b", config.side_effect_re.pattern)
        self.assertNotRegex("", config.side_effect_re)
        self.assertRegex("fizzbuzz", config.side_effect_re)
        self.assertRegex("fizzbuzz.foo", config.side_effect_re)
        self.assertNotRegex("fizzbuzz1", config.side_effect_re)
        self.assertNotRegex("fizzbuzz1.foo", config.side_effect_re)
        self.assertRegex("foo.bar.baz", config.side_effect_re)
        self.assertRegex("foo.bar.baz.qux", config.side_effect_re)
        self.assertNotRegex("foo.bar", config.side_effect_re)
        self.assertNotRegex("foo.bar.qux", config.side_effect_re)

    def test_is_side_effect(self) -> None:
        with self.subTest("fizzbuzz"):
            config = Config(side_effect_modules=["fizzbuzz"])
            # import fizzbuzz1
            self.assertFalse(config.is_side_effect_import("", ["fizzbuzz1"]))
            # import fizzbuzz
            self.assertTrue(config.is_side_effect_import("", ["fizzbuzz"]))
            # from fizzbuzz import a, b
            self.assertTrue(config.is_side_effect_import("fizzbuzz", ["a", "b"]))
            # from fizzbuzz.apple import a, b
            self.assertTrue(config.is_side_effect_import("fizzbuzz.apple", ["a", "b"]))

        with self.subTest("foo.bar.baz"):
            config = Config(side_effect_modules=["foo.bar.baz"])
            # import foo.bar
            self.assertFalse(config.is_side_effect_import("", ["foo.bar"]))
            # import foo, bar
            self.assertFalse(config.is_side_effect_import("", ["foo", "bar"]))
            # import foo.bar.baz
            self.assertTrue(config.is_side_effect_import("", ["foo.bar.baz"]))
            # import foo.bar.bazzy
            self.assertFalse(config.is_side_effect_import("", ["foo.bar.bazzy"]))
            # from foo import bar
            self.assertFalse(config.is_side_effect_import("foo", ["bar"]))
            # from foo.bar import foo
            self.assertFalse(config.is_side_effect_import("foo.bar", ["foo"]))
            # from foo.bar import baz
            self.assertTrue(config.is_side_effect_import("foo.bar", ["baz"]))
            # from foo.bar import bazzy
            self.assertFalse(config.is_side_effect_import("foo.bar", ["bazzy"]))

    def test_config_magic_commas(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            d_path = Path(d)
            (d_path / "foo").mkdir(parents=True)
            (d_path / "foo" / "bar.py").write_text("import os")

            with self.subTest("default"):
                (d_path / "foo" / "pyproject.toml").write_text("")
                conf = Config.find(d_path / "foo" / "bar.py")
                self.assertFalse(conf.magic_commas)

            with self.subTest("enabled"):
                (d_path / "foo" / "pyproject.toml").write_text(
                    """\
[tool.usort]
magic_commas = true
"""
                )
                conf = Config.find(d_path / "foo" / "bar.py")
                self.assertTrue(conf.magic_commas)

    def test_config_merge_imports(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            d_path = Path(d)
            (d_path / "foo").mkdir(parents=True)
            (d_path / "foo" / "bar.py").write_text("import os")

            with self.subTest("default"):
                (d_path / "foo" / "pyproject.toml").write_text("")
                conf = Config.find(d_path / "foo" / "bar.py")
                self.assertTrue(conf.merge_imports)

            with self.subTest("disabled"):
                (d_path / "foo" / "pyproject.toml").write_text(
                    """\
[tool.usort]
merge_imports = false
"""
                )
                conf = Config.find(d_path / "foo" / "bar.py")
                self.assertFalse(conf.merge_imports)

    def test_config_excludes(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            d_path = Path(d)
            (d_path / "foo").mkdir(parents=True)
            (d_path / "foo" / "bar.py").write_text("import os")

            with self.subTest("default config"):
                (d_path / "foo" / "pyproject.toml").write_text("")
                conf = Config.find(d_path / "foo" / "bar.py")
                self.assertEqual([], conf.excludes)

            with self.subTest("with black config"):
                (d_path / "foo" / "pyproject.toml").write_text(
                    """\
[tool.usort]
excludes = [
    "fixtures/",
    "*generated.py",
]
"""
                )
                conf = Config.find(d_path / "foo" / "bar.py")
                expected = ["fixtures/", "*generated.py"]
                self.assertEqual(expected, conf.excludes)

    def test_black_line_length(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            d_path = Path(d)
            (d_path / "foo").mkdir(parents=True)
            (d_path / "foo" / "bar.py").write_text("import os")

            with self.subTest("default config"):
                (d_path / "foo" / "pyproject.toml").write_text("")
                conf = Config.find(d_path / "foo" / "bar.py")
                self.assertEqual(Config.line_length, conf.line_length)

            with self.subTest("with black config"):
                (d_path / "foo" / "pyproject.toml").write_text(
                    """\
[tool.black]
line-length = 120
"""
                )
                conf = Config.find(d_path / "foo" / "bar.py")
                self.assertEqual(120, conf.line_length)

    def test_config_parent_walk(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            d_path = Path(d)

            (d_path / "a").mkdir(parents=True)
            (d_path / "a" / "__init__.py").write_text("")

            (d_path / "b").mkdir(parents=True)
            (d_path / "b" / "pyproject.toml").write_text(
                """\
[tool.usort.known]
first_party = ["x"]
"""
            )
            (d_path / "b" / "c").mkdir(parents=True)
            (d_path / "b" / "c" / "d").symlink_to(d_path / "a")

            # This is behavior that's worked for ages, if given an absolute path
            conf = Config.find(d_path / "b" / "c" / "d" / "__init__.py")
            self.assertEqual(CAT_FIRST_PARTY, conf.known["x"])
            self.assertEqual(CAT_FIRST_PARTY, conf.known["d"])

            # This is also something that's worked for ages, a relative path with cwd
            # above the pyproject.toml
            with chdir(d):
                conf = Config.find(Path("b") / "c" / "d" / "__init__.py")
                self.assertEqual(CAT_FIRST_PARTY, conf.known["x"])
                self.assertEqual(CAT_FIRST_PARTY, conf.known["d"])

            # This is an instance of issue 43
            with chdir((d_path / "b" / "c").as_posix()):
                conf = Config.find(Path("d") / "__init__.py")
                self.assertEqual(CAT_FIRST_PARTY, conf.known["x"])
                self.assertEqual(CAT_FIRST_PARTY, conf.known["d"])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import FrozenSet

from stdlibs import py3

STDLIB_TOP_LEVEL_NAMES: FrozenSet[str] = py3.module_names



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, NewType, Optional, Pattern, Sequence, Set

import toml

from .stdlibs import STDLIB_TOP_LEVEL_NAMES

Category = NewType("Category", str)

CAT_FUTURE = Category("future")
CAT_STANDARD_LIBRARY = Category("standard_library")
CAT_FIRST_PARTY = Category("first_party")
CAT_THIRD_PARTY = Category("third_party")


def known_factory() -> Dict[str, Category]:
    known = {}
    for name in STDLIB_TOP_LEVEL_NAMES:
        known[name] = CAT_STANDARD_LIBRARY

    # This is also in the stdlib list, so this override comes last...
    known["__future__"] = CAT_FUTURE
    known["__main__"] = CAT_FIRST_PARTY

    return known


@dataclass
class Config:
    known: Dict[str, Category] = field(default_factory=known_factory)

    # These names are vaguely for compatibility with isort; however while it has
    # separate sections for "current project" and "explicitly local", these are
    # handled differently as "first_party".
    categories: Sequence[Category] = (
        CAT_FUTURE,
        CAT_STANDARD_LIBRARY,
        CAT_THIRD_PARTY,
        CAT_FIRST_PARTY,
    )
    default_category: Category = CAT_THIRD_PARTY

    # Known set of modules with import side effects. These will be implicitly treated
    # as block separators, similar to non-import statements.
    side_effect_modules: List[str] = field(default_factory=list)
    side_effect_re: Pattern[str] = field(default=re.compile(""))

    # Whether to perform the first-party heuristic during find()
    first_party_detection: bool = True

    # Whether to follow black-style for magic trailing commas
    magic_commas: bool = False

    # Whether to merge imports when sorting
    merge_imports: bool = True

    # gitignore-style filename patterns to exclude when sorting entire directories
    excludes: List[str] = field(default_factory=list)

    # Needed for formatting final imports
    line_length: int = 88

    def __post_init__(self) -> None:
        self.side_effect_re = re.compile(
            "|".join(re.escape(m) + r"\b" for m in self.side_effect_modules)
        )

    @classmethod
    def find(
        cls, filename: Optional[Path] = None, with_first_party: bool = True
    ) -> "Config":
        rv = cls()

        # TODO This logic should be split out to a separate project, as it's
        # reusable and deserves a number of tests to get right.  Can probably
        # also stop once finding a .hg, .git, etc
        if filename is None:
            p = Path.cwd()
        else:
            p = Path.cwd() / filename

        while True:
            if p.is_dir():
                candidate = p / "pyproject.toml"
                if candidate.exists():
                    rv.update_from_config(candidate)
                    break

            # Stop on root (hopefully works on Windows)
            if p.parent == p:
                break
            # Stop on different volume
            if p.exists() and p.stat().st_dev != p.parent.stat().st_dev:
                break

            p = p.parent

        # Either param or config can force off.
        if with_first_party and rv.first_party_detection:
            if filename is None:
                # directory
                rv = rv.with_first_party(Path.cwd())
            else:
                # filename, ideally
                rv = rv.with_first_party(Path.cwd() / filename)

        return rv

    def with_first_party(self, filename: Path) -> "Config":
        """
        Infer first-party top-level names; this only works for the common case of
        a single package, but not multiple, or modules (pure-python or extensions).
        In those cases, you should explicitly set `known_first_party` in the config.

        This code as-is should work for something like running against a site-packages
        dir, but if we broaden to support multiple top-level names, it would find too
        much.

        To disable this code entirely, set `first_party_detection=false` in the config.
        """

        p = filename

        while True:
            # Stop on root (hopefully works on Windows)
            if p.parent == p:
                break
            # Stop on different volume
            if p.exists() and p.stat().st_dev != p.parent.stat().st_dev:
                break

            if (p.parent / "__init__.py").exists():
                p = p.parent
            else:
                break

        if (p / "__init__.py").exists():
            self.known[p.name] = CAT_FIRST_PARTY

        return self

    def update_from_config(self, toml_path: Path) -> None:
        conf = toml.loads(toml_path.read_text())
        tool = conf.get("tool", {})
        tbl = tool.get("usort", {})

        if "categories" in tbl:
            self.categories = [Category(x) for x in tbl["categories"]]
        if "default_category" in tbl:
            self.default_category = Category(tbl["default_category"])
        if "side_effect_modules" in tbl:
            self.side_effect_modules.extend(tbl["side_effect_modules"])
        if "first_party_detection" in tbl:
            self.first_party_detection = tbl["first_party_detection"]
        if "magic_commas" in tbl:
            self.magic_commas = bool(tbl["magic_commas"])
        if "merge_imports" in tbl:
            self.merge_imports = bool(tbl["merge_imports"])
        if "excludes" in tbl:
            self.excludes = tbl["excludes"]

        for cat, names in tbl.get("known", {}).items():
            typed_cat = Category(cat)
            if cat not in self.categories:
                raise ValueError(f"Known set for {cat} without it having an order")

            for name in names:
                self.known[name] = typed_cat

        # "legacy" options
        for cat, option in [
            (CAT_FIRST_PARTY, "known_first_party"),
            (CAT_THIRD_PARTY, "known_third_party"),
            (CAT_STANDARD_LIBRARY, "known_standard_library"),
        ]:
            if option in tbl:
                for name in tbl[option]:
                    # TODO validate (no dots or whitespace, etc)
                    assert "." not in name
                    self.known[name] = cat

        # black config
        tbl = tool.get("black", {})
        if "line-length" in tbl:
            self.line_length = int(tbl["line-length"])

        # make sure generated regexes get updated
        self.__post_init__()

    def category(self, dotted_import: str) -> Category:
        """
        Given a piece of an import string, return its category for this config.

        You can pass in ".foo" or "pkg.foo.bar" or just "os" and it should
        categorize.
        """
        first_part = dotted_import.split(".")[0]
        if first_part == "":
            # relative import
            return CAT_FIRST_PARTY
        elif first_part in self.known:
            return self.known[first_part]
        else:
            return self.default_category

    def is_side_effect_import(self, base: str, names: List[str]) -> bool:
        """
        Determine if any of the given imports are in the list with known side effects.

        Takes a "base" (possibly empty) and a list of imported names, and checks if
        any of the base+name combinations (or a prefix of that combination) is in the
        list of know modules with side effects.
        """
        if self.side_effect_modules:
            candidates: Set[str] = set()
            for name in names:
                candidates.add(f"{base}.{name}" if base else name)
            return any(self.side_effect_re.match(candidate) for candidate in candidates)
        return False



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# noqa: F401

try:
    from .version import version as __version__
except ImportError:
    __version__ = "dev"

from .api import usort, usort_bytes, usort_file, usort_path, usort_stdin, usort_string
from .config import Config
from .types import Result, SortWarning

__all__ = [
    "__version__",
    "Config",
    "Result",
    "SortWarning",
    "usort",
    "usort_bytes",
    "usort_file",
    "usort_path",
    "usort_stdin",
    "usort_string",
]

# DEPRECATED: preserve old api, will be removed by 1.0
from . import sorting  # usort:skip

sorting.usort_bytes = usort_bytes  # type: ignore
sorting.usort_path = usort_path  # type: ignore
sorting.usort_stdin = usort_stdin  # type: ignore
sorting.usort_string = usort_string  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from textwrap import dedent

from .. import types


class TypesTest(unittest.TestCase):
    maxDiff = None

    def test_import_item_comments_add(self) -> None:
        a = types.ImportItemComments(
            before=["# hi"], inline=["# type: ignore"], following=[]
        )
        b = types.ImportItemComments(
            before=[],
            inline=["# noqa"],
            following=["# bye"],
        )
        expected = types.ImportItemComments(
            before=["# hi"],
            inline=["# type: ignore", "# noqa"],
            following=["# bye"],
        )

        with self.subTest("add"):
            self.assertEqual(expected, a + b)

        with self.subTest("inplace"):
            a += b
            self.assertEqual(expected, a)

        with self.subTest("bad types"):
            with self.assertRaisesRegex(
                TypeError, "unsupported.+'ImportItemComments'.+'int'"
            ):
                a += 10  # type: ignore

    def test_import_comments_add(self) -> None:
        a = types.ImportComments(
            before=["# header"],
            first_inline=[],
            initial=[],
            inline=[],
            final=[],
            last_inline=["# footer"],
        )
        b = types.ImportComments(
            before=[],
            first_inline=["# noqa"],
            initial=[],
            inline=[],
            final=[],
            last_inline=["# goodbye"],
        )
        expected = types.ImportComments(
            before=["# header"],
            first_inline=["# noqa"],
            initial=[],
            inline=[],
            final=[],
            last_inline=["# footer", "# goodbye"],
        )

        with self.subTest("add"):
            self.assertEqual(expected, a + b)

        with self.subTest("inplace"):
            a += b
            self.assertEqual(expected, a)

        with self.subTest("bad types"):
            with self.assertRaisesRegex(
                TypeError, "unsupported.+'ImportComments'.+'int'"
            ):
                a += 10  # type: ignore

    def test_sortable_import_item_add(self) -> None:
        a = types.SortableImportItem(
            name="foo",
            asname="foofoo",
            comments=types.ImportItemComments(
                inline=["# noqa"],
            ),
        )
        b = types.SortableImportItem(
            name="foo",
            asname="foofoo",
            comments=types.ImportItemComments(
                before=["# hello"],
            ),
        )
        expected = types.SortableImportItem(
            name="foo",
            asname="foofoo",
            comments=types.ImportItemComments(
                before=["# hello"],
                inline=["# noqa"],
            ),
        )

        with self.subTest("add"):
            self.assertEqual(expected, a + b)

        with self.subTest("inplace"):
            a += b
            self.assertEqual(expected, a)

        with self.subTest("bad types"):
            with self.assertRaisesRegex(
                TypeError, "unsupported.+'SortableImportItem'.+'int'"
            ):
                a += 10  # type: ignore

    def test_sortable_import_add(self) -> None:
        a = types.SortableImport(
            stem="foo",
            items=[
                types.SortableImportItem(
                    name="bar",
                    asname=None,
                    comments=types.ImportItemComments(
                        before=["# hello"],
                    ),
                ),
                types.SortableImportItem(
                    name="baz",
                    asname=None,
                    comments=types.ImportItemComments(
                        inline=["# noqa"],
                    ),
                ),
            ],
            comments=types.ImportComments(
                before=["# original block"],
            ),
            indent="",
        )
        b = types.SortableImport(
            stem="foo",
            items=[
                types.SortableImportItem(
                    name="bar",
                    asname=None,
                    comments=types.ImportItemComments(
                        inline=["# noqa"],
                    ),
                ),
                types.SortableImportItem(
                    name="baz",
                    asname="bazz",
                    comments=types.ImportItemComments(),
                ),
                types.SortableImportItem(
                    name="buzz",
                    asname=None,
                    comments=types.ImportItemComments(),
                ),
            ],
            comments=types.ImportComments(
                before=["# added block"],
            ),
            indent="",
        )
        expected = types.SortableImport(
            stem="foo",
            items=[
                types.SortableImportItem(
                    name="bar",
                    asname=None,
                    comments=types.ImportItemComments(
                        before=["# hello"],
                        inline=["# noqa"],
                    ),
                ),
                types.SortableImportItem(
                    name="baz",
                    asname=None,
                    comments=types.ImportItemComments(
                        inline=["# noqa"],
                    ),
                ),
                types.SortableImportItem(
                    name="baz",
                    asname="bazz",
                    comments=types.ImportItemComments(),
                ),
                types.SortableImportItem(
                    name="buzz",
                    asname=None,
                    comments=types.ImportItemComments(),
                ),
            ],
            comments=types.ImportComments(
                before=["# original block", "# added block"],
            ),
            indent="",
        )

        with self.subTest("add"):
            self.assertEqual(expected, a + b)

        with self.subTest("inplace"):
            a += b
            self.assertEqual(expected, a)

        with self.subTest("bad types"):
            with self.assertRaisesRegex(
                TypeError, "unsupported.+'SortableImport'.+'int'"
            ):
                a += 10  # type: ignore

    def test_sortable_import_trailing_comma(self) -> None:
        imp = types.SortableImport(
            stem="a",
            items=[
                types.SortableImportItem(name="b", asname="", stem="a", comma=True),
                types.SortableImportItem(name="c", asname="", stem="a", comma=True),
                types.SortableImportItem(name="d", asname="", stem="a", comma=False),
            ],
        )
        self.assertFalse(imp.trailing_comma)

        imp = types.SortableImport(
            stem="a",
            items=[
                types.SortableImportItem(name="b", asname="", stem="a", comma=True),
                types.SortableImportItem(name="c", asname="", stem="a", comma=True),
                types.SortableImportItem(name="d", asname="", stem="a", comma=True),
            ],
        )
        self.assertTrue(imp.trailing_comma)

    def test_sortable_block_repr(self) -> None:
        imp = types.SortableBlock(
            start_idx=0,
            end_idx=2,
            imports=[
                types.SortableImport(
                    stem="foo",
                    items=[
                        types.SortableImportItem(
                            "bar", None, types.ImportItemComments()
                        ),
                        types.SortableImportItem(
                            "baz", "buzz", types.ImportItemComments()
                        ),
                    ],
                    comments=types.ImportComments(),
                    indent="    ",
                ),
                types.SortableImport(
                    stem="apple",
                    items=[
                        types.SortableImportItem(
                            "core", None, types.ImportItemComments()
                        ),
                    ],
                    comments=types.ImportComments(),
                    indent="    ",
                ),
            ],
        )

        expected_repr = dedent(
            """
                SortableBlock(
                    start_idx = 0,
                    end_idx = 2,
                    imports = [
                        SortableImport(
                            # sort_key = SortKey(category_index=2, is_from_import=True, ndots=0),
                            stem = 'foo',
                            items = [
                                SortableImportItem(name='bar', asname=None, comments=ImportItemComments(before=[], inline=[], following=[])),
                                SortableImportItem(name='baz', asname='buzz', comments=ImportItemComments(before=[], inline=[], following=[])),
                            ],
                            comments = ImportComments(before=[], first_inline=[], initial=[], inline=[], final=[], last_inline=[]),
                            indent = '    ',
                        ),
                        SortableImport(
                            # sort_key = SortKey(category_index=2, is_from_import=True, ndots=0),
                            stem = 'apple',
                            items = [
                                SortableImportItem(name='core', asname=None, comments=ImportItemComments(before=[], inline=[], following=[])),
                            ],
                            comments = ImportComments(before=[], first_inline=[], initial=[], inline=[], final=[], last_inline=[]),
                            indent = '    ',
                        ),
                    ],
                    imported_names = {},
                )
            """
        ).strip()
        self.assertEqual(expected_repr, repr(imp))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from multiprocessing import freeze_support

from .cli import main

if __name__ == "__main__":
    freeze_support()
    main(prog_name="usort")
